/*
 * Copyright (C) 2012 by
 *   MetraLabs GmbH (MLAB), GERMANY
 * and
 *   Neuroinformatics and Cognitive Robotics Labs (NICR) at TU Ilmenau, GERMANY
 * All rights reserved.
 *
 * Contact: info@mira-project.org
 *
 * Commercial Usage:
 *   Licensees holding valid commercial licenses may use this file in
 *   accordance with the commercial license agreement provided with the
 *   software or, alternatively, in accordance with the terms contained in
 *   a written agreement between you and MLAB or NICR.
 *
 * GNU General Public License Usage:
 *   Alternatively, this file may be used under the terms of the GNU
 *   General Public License version 3.0 as published by the Free Software
 *   Foundation and appearing in the file LICENSE.GPL3 included in the
 *   packaging of this file. Please review the following information to
 *   ensure the GNU General Public License version 3.0 requirements will be
 *   met: http://www.gnu.org/copyleft/gpl.html.
 *   Alternatively you may (at your option) use any later version of the GNU
 *   General Public License if such license has been publicly approved by
 *   MLAB and NICR (or its successors, if any).
 *
 * IN NO EVENT SHALL "MLAB" OR "NICR" BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
 * THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF "MLAB" OR
 * "NICR" HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * "MLAB" AND "NICR" SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND "MLAB" AND "NICR" HAVE NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS.
 */

/**
 * @file Thread.C
 *    Implementation of Thread.h.
 *
 * @author Tim Langner
 * @date   2012/01/18
 */

#include <thread/Thread.h>

#ifdef MIRA_LINUX
# include <sys/times.h>
#elif defined(MIRA_WINDOWS)
# include <windows.h>
# include <error/Exceptions.h>
# include <error/SystemError.h>
#endif

namespace mira {

///////////////////////////////////////////////////////////////////////////////

#ifdef MIRA_LINUX

float getCPUUsage()
{
	static clock_t lastCPU;
	static clock_t lastSystemCPU;
	static clock_t lastUserCPU;
	static bool firstCall = true;
	
	if (firstCall)
	{
		struct tms timeSample;

		lastCPU = times(&timeSample);
		lastSystemCPU = timeSample.tms_stime;
		lastUserCPU = timeSample.tms_utime;
		firstCall = false;
	}
	struct tms timeSample;
	clock_t now;
	float percent;

	now = times(&timeSample);
	if (now <= lastCPU || timeSample.tms_stime < lastSystemCPU ||
	    timeSample.tms_utime < lastUserCPU)
		percent = -1.0f;
	else
	{
		percent = (timeSample.tms_stime - lastSystemCPU) +
		          (timeSample.tms_utime - lastUserCPU);
		percent /= now - lastCPU;
		percent *= 100.0f;
	}
	lastCPU = now;
	lastSystemCPU = timeSample.tms_stime;
	lastUserCPU = timeSample.tms_utime;
	return percent;
}

#elif defined(MIRA_WINDOWS)

float getCPUUsage()
{
	static ULARGE_INTEGER lastCPU;
	static ULARGE_INTEGER lastSystemCPU;
	static ULARGE_INTEGER lastUserCPU;
	static HANDLE self;
	static bool firstCall = true;

	if (firstCall)
	{
		FILETIME ftime, fsys, fuser;

		GetSystemTimeAsFileTime(&ftime);
		memcpy(&lastCPU, &ftime, sizeof(FILETIME));

		self = GetCurrentProcess();
		if (GetProcessTimes(self, &ftime, &ftime, &fsys, &fuser) == 0)
		{
			boost::system::error_code err = getLastSystemErrorCode();
			MIRA_THROW(XSystemCall, "GetProcessTimes in getCPUUsage failed: " <<
			           err.message());
		}
		memcpy(&lastSystemCPU, &fsys, sizeof(FILETIME));
		memcpy(&lastUserCPU, &fuser, sizeof(FILETIME));
		firstCall = false;
	}

	FILETIME ftime, fsys, fuser;
	ULARGE_INTEGER now, sys, user;
	float percent;

	GetSystemTimeAsFileTime(&ftime);
	memcpy(&now, &ftime, sizeof(FILETIME));

	if (GetProcessTimes(self, &ftime, &ftime, &fsys, &fuser) == 0)
	{
		boost::system::error_code err = getLastSystemErrorCode();
		MIRA_THROW(XSystemCall, "GetProcessTimes in getCPUUsage failed: " <<
		           err.message());
	}
	memcpy(&sys, &fsys, sizeof(FILETIME));
	memcpy(&user, &fuser, sizeof(FILETIME));
	percent = (sys.QuadPart - lastSystemCPU.QuadPart) +
			(user.QuadPart - lastUserCPU.QuadPart);
	percent /= (now.QuadPart - lastCPU.QuadPart);
	lastCPU = now;
	lastUserCPU = user;
	lastSystemCPU = sys;

	return percent * 100.0f;
}

#endif

///////////////////////////////////////////////////////////////////////////////

}
