#!/bin/bash

# we need miracd from mirabash, so import that file
DIR=$(dirname "${BASH_SOURCE[0]}")
source $DIR/../mirabash

LC_ALL=C

# exit on Ctrl+C
trap ctrl_c INT
function ctrl_c() {
	exit 2
}


function usage
{
	echo -e "\
Usage: miragit <command> [<package> [<package> ...]] [options]\n\
Runs 'git' for the specified packages or for all packages (if no package is specified).\n\n\
General Options:\n\
  --at <location>: only process packages in a certain location (directory). Can be used with absolute or relative paths (also e.g. '.')
  --auth: interactively query username and email and provide to git (useful for committing on a machine where these are not global settings)
\n\
Available commands:\n\
    status - status of working tree\n\
    commit - record changes to repository(use with care!)\n\
    push   - update remote refs(use with care!)\n\
    pull   - fetch from and integrate with remote repository\n\
\n\
    all other git commands are supported too, type 'git help <command>' for details.\n"
}

function gitpackage
{
	miracd $1 &> /dev/null
	if [[ $? != 0 ]]; then
		echo -e "\e[01;31mPackage '$1' No such package found\e[00m"
	else
		here=`realpath $PWD`
		if [[ $here"/" != "$4"/* ]]; then
			return
		fi

		### check if we are in a git directory
		git status 2>&1 1>&/dev/null; ret=$?

		if [[ $ret == 0 ]]; then

			echo -e "\e[01;32mPackage '$1' \e[00;32m[$PWD]\e[00m"
			
			IFS_BACKUP=$IFS
			IFS=$(echo -en "\n\b")
			if [[ -z "$5" ]]; then
				OUTPUT=`git $2 $3`
			else
				OUTPUT=`git -c user.name="$5" -c user.email=$6 $2 $3`
			fi
			if [[ $? != 0 ]]; then
				GIT_ERROR=1
			fi
			if [[ $2 = "pull" ]]; then
				echo "$OUTPUT" | egrep --invert-match -e "Already up to date." --invert-match -e "Already up-to-date."
			elif [[ $2 = "push" ]]; then
				echo "$OUTPUT" | egrep --invert-match -e "Everything up to date." --invert-match -e "Everything up-to-date."
			else
				echo "$OUTPUT"
			fi
			IFS=$IFS_BACKUP
		fi
	fi
}

if [[ $1 = "--help" ]] || [[ $# -lt 1 ]]; then
	usage
	exit 0
fi

# if there is no credentials helper configured, enable cache, so user does not need to enter password for each package
# backup the git config and restore it on exit
configfile=${HOME}/.gitconfig
configfilebak=${HOME}/.gitconfig.miragit
configfile_backedup=0
credentialscache_enabled=0

if [[ -f ${configfile} ]]; then
	#echo "config exists - backing up"
	mv ${configfile} ${configfilebak} # keep original timestamp
	cp ${configfilebak} ${configfile}
	configfile_backedup=1
fi

credential_helper="$(git config --get credential.helper)"
ret=$?
if [[ $ret -ne 0 ]] || [[ credential_helper = "" ]]; then
	echo "No existing credential helper found - using cache"
	credentialscache_enabled=1
	git config --global credential.helper cache
fi

# first param is the command
command="$1"

# other params may be packages, the others are the options for git
# find the first parameter, that does contain a -, it will be the first option to 
# pass to git
paramcount=$#
firstoption=$((${paramcount}+1))

for (( i=2; i<=${paramcount}; i++ ))
do
	if [[ ${!i} == -* ]]; then
		firstoption=$i
		break
	fi
done

firstgitoption=$firstoption

username=""
email=""
at=""

while :
do
	if [[ ${!firstgitoption} == "--auth" ]]; then
		read -p "user.name (Firstname Secondname): " username
		read -p "user.email: " email
		echo ""
		firstgitoption=$(($firstgitoption+1))
	elif [[ ${!firstgitoption} == "--at" ]]; then
		firstgitoption=$(($firstgitoption+1))
		at=`realpath ${!firstgitoption}`
		firstgitoption=$(($firstgitoption+1))
		echo "Restrict to packages in directory" $at
	else
		break
	fi
done

gitoptions="${@:${firstgitoption}}"

# handle additional commands
if [[ $command == "changes" ]]; then
	command="log -r BASE:HEAD"
fi

if [[ $firstoption == 2 ]]; then
	echo "Running 'git $command $gitoptions' for all packages ..."
	echo "Reindexing packages ..." 
	$_mirafindpkg --reindex
	allpackages=($(bash $_mirafindpkg | awk -- '{print $1}'))
	GIT_ERROR=0
	for package in "${allpackages[@]}"	
	do
		gitpackage $package "$command" "$gitoptions" "$at" "$username" $email
	done
	if [[ $GIT_ERROR != 0 ]]; then
		echo -e "\e[00;31mThere were errors reported by git. Make sure to check output above!\e[00m"
	fi
else 
	# for each given package
	echo "Running 'git $command $gitoptions' ..."
	GIT_ERROR=0
	for (( i=2; i<${firstoption}; i++ ))
	do
		package="${!i}"
		gitpackage $package "$command" "$gitoptions" "$at" "$username" $email
	done
	if [[ $GIT_ERROR != 0 ]]; then
		echo -e "\e[00;31mThere were errors reported by git. Make sure to check output above!\e[00m"
	fi
fi

# if this script enabled credential caching, disable it now
if [[ ${credentialscache_enabled} -eq 1 ]]; then
	echo "Disabling credentials cache again"
	git credential-cache exit
fi

rm ${configfile}
if [[ ${configfile_backedup} -eq 1 ]]; then
	#echo "restoring config"
	mv ${configfilebak} ${configfile} 
fi

